# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.errors import *
from montepy.surfaces.surface_type import SurfaceType
from montepy.surfaces.surface import Surface


class GeneralPlane(Surface):
    """
    Represents P

    :param input: The Input object representing the input
    :type input: Input
    """

    def __init__(self, input=None):
        super().__init__(input)
        if input:
            if self.surface_type != SurfaceType.P:
                raise ValueError("A GeneralPlane must be a surface of type P")
            if len(self.surface_constants) not in {4, 9}:
                raise ValueError(
                    "A GeneralPlane must have either 4 or 9 surface constants"
                )

    def validate(self):
        super().validate()
        if len(self.surface_constants) not in {4, 9}:
            raise IllegalState(
                f"Surface: {self.number} does not have constants set properly."
            )



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.numbered_object_collection import NumberedDataObjectCollection
from montepy.data_inputs.transform import Transform


class Transforms(NumberedDataObjectCollection):
    """
    A container of multiple :class:`~montepy.data_inputs.transform.Transform` instances.
    """

    def __init__(self, objects=None, problem=None):
        super().__init__(Transform, objects, problem)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.data_inputs.data_input import DataInputAbstract
from montepy.input_parser import syntax_node
from montepy.particle import Particle


class Mode(DataInputAbstract):
    """
    Class for the particle mode for a problem.

    :param input: the Input object representing this data input
    :type input: Input
    """

    def __init__(self, input=None):
        super().__init__(input)
        if input:
            self._particles = set()
            self._parse_and_override_particle_modes(
                [p.value for p in self._tree["data"]]
            )
        else:
            self._particles = {Particle.NEUTRON}
            classifier = syntax_node.ClassifierNode()
            classifier.prefix = self._generate_default_node(str, "MODE")
            classifier.padding = syntax_node.PaddingNode(" ")
            self._tree = syntax_node.SyntaxNode(
                "mode",
                {
                    "classifier": classifier,
                    "data": syntax_node.ListNode("particles"),
                },
            )

    def _parse_and_override_particle_modes(self, particles):
        self._particles = set()
        for particle in particles:
            if not isinstance(particle, str):
                raise TypeError(f"Mode particle must be a str. {particle} given.")
            self._particles.add(Particle(particle.upper()))

    @property
    def particles(self):
        """
        The type of particles involved in this problem.

        The set will contain instances of :class:`montepy.particle.Particle`.

        :rtype: set
        """
        return self._particles.copy()

    def add(self, particle):
        """
        Adds the given particle to the problem.

        If specifying particle type by string this must be the MCNP shorthand,
        such as ``n`` for ``Particle.NEUTRON``.

        :param particle: the particle type to add to the mode.
        :type particle: Particle, str
        :raises ValueError: if string is not a valid particle shorthand.
        """
        if not isinstance(particle, (Particle, str, syntax_node.ValueNode)):
            raise TypeError("particle must be a Particle instance")
        if isinstance(particle, (str, syntax_node.ValueNode)):
            # error catching not needed
            # enum will raise ValueError "foo is not a valid Particle"
            if isinstance(particle, syntax_node.ValueNode):
                particle = particle.value
            particle = Particle(particle.upper())
        self._particles.add(particle)

    def remove(self, particle):
        """
        Remove the given particle from the problem

        :param particle: the particle type to remove from the mode.
        :type particle: Particle, str
        :raises ValueError: if string is not a valid particle shorthand.
        """
        if not isinstance(particle, (Particle, str)):
            raise TypeError("particle must be a Particle instance")
        if isinstance(particle, str):
            particle = Particle(particle.upper())
        self._particles.remove(particle)

    def set(self, particles):
        """
        Completely override the current mode.

        Can specify it as:
         * ``"n p"``
         * ``["n", "p"]``
         * ``[Particle.NEUTRON, Particle.PHOTON]``

        :param particles: the particles that the mode will be switched to.
        :type particles: list, str
        :raises ValueError: if string is not a valid particle shorthand.
        """
        if not isinstance(particles, (list, set, str)):
            raise TypeError("particles must be a list, string, or set")
        if isinstance(particles, (list, set)):
            is_str = True
            for particle in particles:
                if not isinstance(particle, (str, Particle)):
                    raise TypeError("particle must be a Particle or string")
                if not isinstance(particle, str):
                    is_str = False
        else:
            particles = particles.split()
            is_str = True
        if is_str:
            self._parse_and_override_particle_modes(particles)
        else:
            for particle in particles:
                if not isinstance(particle, Particle):
                    raise ValueError("cannot mix particle and string in mode")
            self._particles = set(particles)

    def __contains__(self, obj):
        return obj in self._particles

    def __iter__(self):
        return iter(self._particles)

    def __len__(self):
        return len(self._particles)

    def __str__(self):
        return f"Mode: {self.particles}"

    @staticmethod
    def _class_prefix():
        return "mode"

    @staticmethod
    def _has_number():
        return False

    @staticmethod
    def _has_classifier():
        return 0

    def _update_values(self):
        old_particles = {}
        for particle in self._tree["data"]:
            part = Particle(particle.value.upper())
            old_particles[part] = particle
        old_parts = set(old_particles.keys())
        to_remove = old_parts - self.particles
        to_add = self.particles - old_parts
        for removal in to_remove:
            node = old_particles[removal]
            self._tree["data"].remove(node)
        for addition in to_add:
            self._tree["data"].append(self._generate_default_node(str, addition.value))



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
from montepy.data_inputs import data_input, thermal_scattering
from montepy.data_inputs.isotope import Isotope
from montepy.data_inputs.material_component import MaterialComponent
from montepy.input_parser import syntax_node
from montepy.input_parser.material_parser import MaterialParser
from montepy import mcnp_object
from montepy.numbered_mcnp_object import Numbered_MCNP_Object
from montepy.errors import *
from montepy.utilities import *
import itertools
import re

import warnings


class Material(data_input.DataInputAbstract, Numbered_MCNP_Object):
    """
    A class to represent an MCNP material.

    .. note::

        There is a known bug (:issue:`182`) that valid MCNP material definitions cannot be parsed.


    :param input: the input card that contains the data
    :type input: Input
    """

    _parser = MaterialParser()

    def __init__(self, input=None):
        self._material_components = {}
        self._thermal_scattering = None
        self._is_atom_fraction = True
        self._number = self._generate_default_node(int, -1)
        super().__init__(input)
        if input:
            num = self._input_number
            self._old_number = copy.deepcopy(num)
            self._number = num
            set_atom_frac = False
            isotope_fractions = self._tree["data"]
            if isinstance(isotope_fractions, syntax_node.IsotopesNode):
                iterator = iter(isotope_fractions)
            else:  # pragma: no cover
                # this is a fall through error, that should never be raised,
                # but is here just in case
                raise MalformedInputError(
                    input,
                    f"Material definitions for material: {self.number} is not valid.",
                )
            for isotope_node, fraction in iterator:
                isotope = Isotope(node=isotope_node, suppress_warning=True)
                fraction.is_negatable_float = True
                if not set_atom_frac:
                    set_atom_frac = True
                    if not fraction.is_negative:
                        self._is_atom_fraction = True
                    else:
                        self._is_atom_fraction = False
                else:
                    # if switching fraction formatting
                    if (not fraction.is_negative and not self._is_atom_fraction) or (
                        fraction.is_negative and self._is_atom_fraction
                    ):
                        raise MalformedInputError(
                            input,
                            f"Material definitions for material: {self.number} cannot use atom and mass fraction at the same time",
                        )

                self._material_components[isotope] = MaterialComponent(
                    isotope, fraction, suppress_warning=True
                )

    @make_prop_val_node("_old_number")
    def old_number(self):
        """
        The material number that was used in the read file

        :rtype: int
        """
        pass

    @property
    def is_atom_fraction(self):
        """
        If true this constituent is in atom fraction, not weight fraction.

        :rtype: bool
        """
        return self._is_atom_fraction

    @property
    def material_components(self):
        """
            The internal dictionary containing all the components of this material.

        .. deprecated:: 0.4.1
            MaterialComponent has been deprecated as part of a redesign for the material
            interface due to a critical bug in how MontePy handles duplicate nuclides.
            See :ref:`migrate 0 1`.

            The keys are :class:`~montepy.data_inputs.isotope.Isotope` instances, and the values are
            :class:`~montepy.data_inputs.material_component.MaterialComponent` instances.

            :rtype: dict
        """
        warnings.warn(
            f"""material_components is deprecated, and will be removed in MontePy 1.0.0.
See <https://www.montepy.org/migrations/migrate0_1.html> for more information """,
            DeprecationWarning,
        )
        return self._material_components

    @make_prop_pointer("_thermal_scattering", thermal_scattering.ThermalScatteringLaw)
    def thermal_scattering(self):
        """
        The thermal scattering law for this material

        :rtype: ThermalScatteringLaw
        """
        return self._thermal_scattering

    @property
    def cells(self):
        """A generator of the cells that use this material.

        :returns: an iterator of the Cell objects which use this.
        :rtype: generator
        """
        if self._problem:
            for cell in self._problem.cells:
                if cell.material == self:
                    yield cell

    def format_for_mcnp_input(self, mcnp_version):
        """
        Creates a string representation of this MCNP_Object that can be
        written to file.

        :param mcnp_version: The tuple for the MCNP version that must be exported to.
        :type mcnp_version: tuple
        :return: a list of strings for the lines that this input will occupy.
        :rtype: list
        """
        lines = super().format_for_mcnp_input(mcnp_version)
        if self.thermal_scattering is not None:
            lines += self.thermal_scattering.format_for_mcnp_input(mcnp_version)
        return lines

    def _update_values(self):
        new_list = syntax_node.IsotopesNode("new isotope list")
        for isotope, component in self._material_components.items():
            isotope._tree.value = isotope.mcnp_str()
            new_list.append(("_", isotope._tree, component._tree))
        self._tree.nodes["data"] = new_list

    def add_thermal_scattering(self, law):
        """
        Adds thermal scattering law to the material

        :param law: the law that is mcnp formatted
        :type law: str
        """
        if not isinstance(law, str):
            raise TypeError(
                f"Thermal Scattering law for material {self.number} must be a string"
            )
        self._thermal_scattering = thermal_scattering.ThermalScatteringLaw(
            material=self
        )
        self._thermal_scattering.add_scattering_law(law)

    def update_pointers(self, data_inputs):
        """
        Updates pointer to the thermal scattering data

        :param data_inputs: a list of the data inputs in the problem
        :type data_inputs: list
        """
        pass

    @staticmethod
    def _class_prefix():
        return "m"

    @staticmethod
    def _has_number():
        return True

    @staticmethod
    def _has_classifier():
        return 0

    def __repr__(self):
        ret = f"MATERIAL: {self.number} fractions: "
        if self.is_atom_fraction:
            ret += "atom\n"
        else:
            ret += "mass\n"

        for component in self._material_components:
            ret += repr(self._material_components[component]) + "\n"
        if self.thermal_scattering:
            ret += f"Thermal Scattering: {self.thermal_scattering}"

        return ret

    def __str__(self):
        elements = self._get_material_elements()
        return f"MATERIAL: {self.number}, {elements}"

    def _get_material_elements(self):
        sortable_components = [
            (iso, component.fraction)
            for iso, component in self._material_components.items()
        ]
        sorted_comps = sorted(sortable_components)
        elements_set = set()
        elements = []
        for isotope, _ in sorted_comps:
            if isotope.element not in elements_set:
                elements_set.add(isotope.element)
                elements.append(isotope.element.name)
        return elements

    def validate(self):
        if len(self._material_components) == 0:
            raise IllegalState(
                f"Material: {self.number} does not have any components defined."
            )

    def __hash__(self):
        """WARNING: this is a temporary solution to make sets remove duplicate materials.

        This should be fixed in the future to avoid issues with object mutation:
            <https://eng.lyft.com/hashing-and-equality-in-python-2ea8c738fb9d>

        """
        temp_hash = ""
        sorted_isotopes = sorted(list(self._material_components.keys()))
        for isotope in sorted_isotopes:
            temp_hash = hash(
                (temp_hash, str(isotope), self._material_components[isotope].fraction)
            )

        return hash((temp_hash, self.number))

    def __eq__(self, other):
        return hash(self) == hash(other)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import itertools

from montepy.input_parser.data_parser import DataParser
from montepy.input_parser import syntax_node


class MaterialParser(DataParser):
    debugfile = None

    @_(
        "introduction isotopes",
        "introduction isotopes parameters",
        "introduction isotopes mat_parameters",
    )
    def material(self, p):
        ret = {}
        for key, node in p.introduction.nodes.items():
            ret[key] = node
        ret["data"] = p.isotopes
        if len(p) > 2:
            ret["parameters"] = p[2]
        return syntax_node.SyntaxNode("data", ret)

    @_("isotope_fractions", "number_sequence", "isotope_hybrid_fractions")
    def isotopes(self, p):
        if hasattr(p, "number_sequence"):
            return self._convert_to_isotope(p.number_sequence)
        return p[0]

    @_("number_sequence isotope_fraction", "isotope_hybrid_fractions isotope_fraction")
    def isotope_hybrid_fractions(self, p):
        if hasattr(p, "number_sequence"):
            ret = self._convert_to_isotope(p.number_sequence)
        else:
            ret = p[0]
        ret.append(p.isotope_fraction)
        return ret

    def _convert_to_isotope(self, old):
        new_list = syntax_node.IsotopesNode("converted isotopes")

        def batch_gen():
            it = iter(old)
            while batch := tuple(itertools.islice(it, 2)):
                yield batch

        for group in batch_gen():
            new_list.append(("foo", *group))
        return new_list

    @_(
        "mat_parameter",
        "parameter",
        "mat_parameters mat_parameter",
        "mat_parameters parameter",
    )
    def mat_parameters(self, p):
        """
        A list of the parameters (key, value pairs) that allows material libraries.

        :returns: all parameters
        :rtype: ParametersNode
        """
        if len(p) == 1:
            params = syntax_node.ParametersNode()
            param = p[0]
        else:
            params = p[0]
            param = p[1]
        params.append(param)
        return params

    @_(
        "classifier param_seperator library",
    )
    def mat_parameter(self, p):
        """
        A singular Key-value pair that includes a material library.

        :returns: the parameter.
        :rtype: SyntaxNode
        """
        return syntax_node.SyntaxNode(
            p.classifier.prefix.value,
            {"classifier": p.classifier, "seperator": p.param_seperator, "data": p[2]},
        )

    @_("NUMBER_WORD")
    @_("NUMBER_WORD padding")
    def library(self, p):
        """
        A library name.
        """
        return self._flush_phrase(p, str)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
from enum import Enum
import itertools
import os
import warnings

from montepy.data_inputs import mode, transform
from montepy._cell_data_control import CellDataPrintController
from montepy.cell import Cell
from montepy.cells import Cells
from montepy.errors import *
from montepy.constants import DEFAULT_VERSION
from montepy.materials import Material, Materials
from montepy.surfaces import surface, surface_builder
from montepy.surface_collection import Surfaces

# weird way to avoid circular imports
from montepy.data_inputs import parse_data
from montepy.input_parser import input_syntax_reader, block_type, mcnp_input
from montepy.input_parser.input_file import MCNP_InputFile
from montepy.universes import Universes
from montepy.transforms import Transforms
import montepy


class MCNP_Problem:
    """
    A class to represent an entire MCNP problem in a semantic way.

    .. note::
        If a stream is provided. It will not be closed by this function.

    :param destination: the path to the input file to read, or a readable stream.
    :type destination: io.TextIOBase, str, os.PathLike
    """

    _NUMBERED_OBJ_MAP = {
        Cell: Cells,
        surface.Surface: Surfaces,
        Material: Materials,
        transform.Transform: Transforms,
        montepy.universe.Universe: Universes,
    }

    def __init__(self, destination):
        if hasattr(destination, "read") and callable(getattr(destination, "read")):
            self._input_file = MCNP_InputFile.from_open_stream(destination)
        elif isinstance(destination, (str, os.PathLike)):
            self._input_file = MCNP_InputFile(destination)
        self._title = None
        self._message = None
        self.__unpickled = False
        self._print_in_data_block = CellDataPrintController()
        self._original_inputs = []
        for collect_type in self._NUMBERED_OBJ_MAP.values():
            attr_name = f"_{collect_type.__name__.lower()}"
            setattr(self, attr_name, collect_type(problem=self))
        self._data_inputs = []
        self._mcnp_version = DEFAULT_VERSION
        self._mode = mode.Mode()

    def __setstate__(self, nom_nom):
        self.__dict__.update(nom_nom)
        self.__unpickled = True

    @staticmethod
    def __get_collect_attr_name(collect_type):
        return f"_{collect_type.__name__.lower()}"

    @property
    def original_inputs(self):
        """
        A list of the MCNP_Inputs read from the original file.

        This should not be mutated, and should be used as a reference to maintain
        the structure

        .. deprecated:: 0.2.0
            This will likely be removed soon, and it's functionality will not be necessary to reproduce.

        :return: A list of the MCNP_Object objects representing the file as it was read
        :rtype: list
        """
        return self._original_inputs

    def __relink_objs(self):
        if self.__unpickled:
            for collection in set(self._NUMBERED_OBJ_MAP.values()) | {"_data_inputs"}:
                if not isinstance(collection, str):
                    collection = self.__get_collect_attr_name(collection)
                collection = getattr(self, collection)
                if isinstance(
                    collection,
                    montepy.numbered_object_collection.NumberedObjectCollection,
                ):
                    collection.link_to_problem(self)
                else:
                    for obj in collection:
                        obj.link_to_problem(self)
            self.__unpickled = False

    def __unlink_objs(self):
        for collection in set(self._NUMBERED_OBJ_MAP.values()) | {"_data_inputs"}:
            if not isinstance(collection, str):
                collection = self.__get_collect_attr_name(collection)
            collection = getattr(self, collection)
            if isinstance(
                collection,
                montepy.numbered_object_collection.NumberedObjectCollection,
            ):
                collection.link_to_problem(None)
            else:
                for obj in collection:
                    obj.link_to_problem(None)
        self.__unpickled = True
        self.__relink_objs()

    def __deepcopy__(self, memo):
        cls = type(self)
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        result.__unlink_objs()
        return result

    def clone(self):
        """
        Creates a complete independent copy of this problem.

        .. versionadded:: 0.5.0

        :rtype: MCNP_Problem
        """
        return copy.deepcopy(self)

    @property
    def cells(self):
        """
        A collection of the Cell objects in this problem.

        :return: a collection of the Cell objects, ordered by the order they were in the input file.
        :rtype: Cells
        """
        self.__relink_objs()
        return self._cells

    @cells.setter
    def cells(self, cells):
        if not isinstance(cells, (Cells, list)):
            raise TypeError("cells must be an instance of list or Cells")
        if isinstance(cells, list):
            cells = Cells(cells)
        if cells is self.cells:
            return
        self.cells.clear()
        self.cells.extend(cells)

    @property
    def mode(self):
        """
        The mode of particles being used for the problem.

        :rtype: Mode
        """
        return self._mode

    def set_mode(self, particles):
        """Sets the mode of problem to the given particles.

        For details see: :func:`montepy.data_cards.mode.Mode.set`.

        :param particles: the particles that the mode will be switched to.
        :type particles: list, str
        :raises ValueError: if string is not a valid particle shorthand.
        """
        self._mode.set(particles)

    @property
    def mcnp_version(self):
        """
        The version of MCNP that this is intended for.

        .. note::
            MCNP versions prior to 6.2 aren't fully supported to avoid
            Export Control Restrictions. Documentation for MCNP 6.2 is public in report:
            LA-UR-17-29981.
            All features are based on MCNP 6.2, and may cause other versions of MCNP to break.

        The version is a tuple of major, minor, revision.
        6.2.0 would be represented as (6, 2, 0)

        :rtype: tuple
        """
        return self._mcnp_version

    @mcnp_version.setter
    def mcnp_version(self, version):
        """
        :param version: the version tuple. Must be greater than 6.2.0
        :type version: tuple
        """
        if version < (5, 1, 60):
            raise ValueError(f"The mcnp_version {version} is not supported by MontePy")
        self._mcnp_version = version

    @property
    def surfaces(self):
        """
        A collection of the Surface objects in this problem.

        :return: a collection of the Surface objects, ordered by the order they were in the input file.
        :rtype: Surfaces
        """
        self.__relink_objs()
        return self._surfaces

    @surfaces.setter
    def surfaces(self, surfs):
        if not isinstance(surfs, (list, Surfaces)):
            raise TypeError("Surfaces must be of type list or Surfaces")
        if isinstance(surfs, list):
            surfs = Surfaces(surfs)
        surfs.link_to_problem(self)
        self._surfaces = surfs

    @property
    def materials(self):
        """
        A collection of the Material objects in this problem.

        :return: a colection of the Material objects, ordered by the order they were in the input file.
        :rtype: Materials
        """
        self.__relink_objs()
        return self._materials

    @materials.setter
    def materials(self, mats):
        if not isinstance(mats, (list, Materials)):
            raise TypeError("materials must be of type list and Materials")
        if isinstance(mats, list):
            mats = Materials(mats)
        mats.link_to_problem(self)
        self._materials = mats

    @property
    def print_in_data_block(self):
        """
        Controls whether or not the specific input gets printed in the cell block or the data block.

        This acts like a dictionary. The key is the case insensitive name of the card.
        For example to enable printing importance data in the data block run:

        ``problem.print_in_data_block["Imp"] = True``

        :rtype: bool
        """
        return self._print_in_data_block

    @property
    def data_inputs(self):
        """
        A list of the DataInput objects in this problem.

        :return: a list of the :class:`~montepy.data_cards.data_card.DataCardAbstract` objects, ordered by the order they were in the input file.
        :rtype: list
        """
        self.__relink_objs()
        return self._data_inputs

    @property
    def input_file(self):
        """
        The file name of the original file name this problem was read from.

        :rtype: MCNP_InputFile
        """
        return self._input_file

    @property
    def message(self):
        """
        The Message object at the beginning of the problem if any.

        :rtype: Message
        """
        return self._message

    @property
    def title(self):
        """
        The Title object for the title.

        :rtype: Title
        """
        return self._title

    @title.setter
    def title(self, title):
        """
        :type title: The str for the title to be set to.
        """
        self._title = mcnp_input.Title([title], title)

    @property
    def universes(self):
        """
        The Universes object holding all problem universes.

        :returns: a collection of universes in the problem.
        :rtype: Universes
        """
        return self._universes

    @property
    def transforms(self):
        """
        The collection of transform objects in this problem.

        :returns: a collection of transforms in the problem.
        :rtype: Transforms
        """
        return self._transforms

    def parse_input(self, check_input=False, replace=True):
        """
        Semantically parses the MCNP file provided to the constructor.

        :param check_input: If true, will try to find all errors with input and collect them as warnings to log.
        :type check_input: bool
        :param replace: replace all non-ASCII characters with a space (0x20)
        :type replace: bool
        """
        trailing_comment = None
        last_obj = None
        last_block = None
        OBJ_MATCHER = {
            block_type.BlockType.CELL: (Cell, self._cells),
            block_type.BlockType.SURFACE: (
                surface_builder.surface_builder,
                self._surfaces,
            ),
            block_type.BlockType.DATA: (parse_data, self._data_inputs),
        }
        try:
            for i, input in enumerate(
                input_syntax_reader.read_input_syntax(
                    self._input_file, self.mcnp_version, replace=replace
                )
            ):
                self._original_inputs.append(input)
                if i == 0 and isinstance(input, mcnp_input.Message):
                    self._message = input

                elif isinstance(input, mcnp_input.Title) and self._title is None:
                    self._title = input

                elif isinstance(input, mcnp_input.Input):
                    if last_block != input.block_type:
                        trailing_comment = None
                        last_block = input.block_type
                    obj_parser, obj_container = OBJ_MATCHER[input.block_type]
                    if len(input.input_lines) > 0:
                        try:
                            obj = obj_parser(input)
                            obj.link_to_problem(self)
                            obj_container.append(obj)
                        except (
                            MalformedInputError,
                            NumberConflictError,
                            ParsingError,
                            UnknownElement,
                        ) as e:
                            if check_input:
                                warnings.warn(
                                    f"{type(e).__name__}: {e.message}", stacklevel=2
                                )
                                continue
                            else:
                                raise e
                        if isinstance(obj, Material):
                            self._materials.append(obj, False)
                        if isinstance(obj, transform.Transform):
                            self._transforms.append(obj, False)
                    if trailing_comment is not None and last_obj is not None:
                        obj._grab_beginning_comment(trailing_comment, last_obj)
                        last_obj._delete_trailing_comment()
                    trailing_comment = obj.trailing_comment
                    last_obj = obj
        except UnsupportedFeature as e:
            if check_input:
                warnings.warn(f"{type(e).__name__}: {e.message}", stacklevel=2)
            else:
                raise e
        self.__update_internal_pointers(check_input)

    def __update_internal_pointers(self, check_input=False):
        """Updates the internal pointers between objects

        :param check_input: If true, will try to find all errors with input and collect them as warnings to log.
        :type check_input: bool
        """

        def handle_error(e):
            if check_input:
                warnings.warn(f"{type(e).__name__}: {e.message}", stacklevel=3)
            else:
                raise e

        self.__load_data_inputs_to_object(self._data_inputs)
        self._cells.update_pointers(
            self.cells,
            self.materials,
            self.surfaces,
            self._data_inputs,
            self,
            check_input,
        )
        for surface in self._surfaces:
            try:
                surface.update_pointers(self.surfaces, self._data_inputs)
            except (
                BrokenObjectLinkError,
                ParticleTypeNotInProblem,
                ParticleTypeNotInCell,
            ) as e:
                handle_error(e)
        to_delete = []
        for data_index, data_input in enumerate(self._data_inputs):
            try:
                if data_input.update_pointers(self._data_inputs):
                    to_delete.append(data_index)
            except (
                BrokenObjectLinkError,
                MalformedInputError,
                ParticleTypeNotInProblem,
                ParticleTypeNotInCell,
            ) as e:
                handle_error(e)
                continue
        for delete_index in to_delete[::-1]:
            del self._data_inputs[delete_index]

    def remove_duplicate_surfaces(self, tolerance):
        """Finds duplicate surfaces in the problem, and remove them.

        :param tolerance: The amount of relative error to consider two surfaces identical
        :type tolerance: float
        """
        to_delete = set()
        matching_map = {}
        for surface in self.surfaces:
            if surface not in to_delete:
                matches = surface.find_duplicate_surfaces(self.surfaces, tolerance)
                if matches:
                    for match in matches:
                        to_delete.add(match)
                        matching_map[match] = surface
        for cell in self.cells:
            cell.remove_duplicate_surfaces(matching_map)
        self.__update_internal_pointers()
        for surface in to_delete:
            self._surfaces.remove(surface)

    def add_cell_children_to_problem(self):
        """
        Adds the surfaces, materials, and transforms of all cells in this problem to this problem to the
        internal lists to allow them to be written to file.

        .. warning::
            this does not move complement cells, and probably other objects.
        """
        surfaces = set(self.surfaces)
        materials = set(self.materials)
        transforms = set(self.transforms)
        for cell in self.cells:
            surfaces.update(set(cell.surfaces))
            for surf in cell.surfaces:
                if surf.transform:
                    transforms.add(surf.transform)
            if cell.material:
                materials.add(cell.material)
        surfaces = sorted(surfaces)
        materials = sorted(materials)
        transforms = sorted(transforms)
        self._surfaces = Surfaces(surfaces, problem=self)
        self._materials = Materials(materials, problem=self)
        self._transforms = Transforms(transforms, problem=self)
        self._data_inputs = sorted(set(self._data_inputs + materials + transforms))

    def write_problem(self, destination, overwrite=False):
        """
        Write the problem to a file or writeable object.

        :param destination: File path or writable object
        :type destination: io.TextIOBase, str, os.PathLike
        :param overwrite: Whether to overwrite 'destination' if it is an existing file
        :type overwrite: bool
        """
        if hasattr(destination, "write") and callable(getattr(destination, "write")):
            new_file = MCNP_InputFile.from_open_stream(destination)
            self._write_to_stream(new_file)
        elif isinstance(destination, (str, os.PathLike)):
            new_file = MCNP_InputFile(destination, overwrite=overwrite)
            with new_file.open("w") as fh:
                self._write_to_stream(fh)
        else:
            raise TypeError(
                f"destination f{destination} is not a file path or writable object"
            )

    def write_to_file(self, file_path, overwrite=False):
        """
        Writes the problem to a file.

        .. versionchanged:: 0.3.0
            The overwrite parameter was added.

        :param file_path: the file path to write this problem to
        :type file_path: str, os.PathLike
        :param overwrite: Whether to overwrite the file at 'new_problem' if it exists
        :type overwrite: bool
        :raises IllegalState: if an object in the problem has not been fully initialized.
        :raises FileExistsError: if a file already exists with the same path.
        :raises IsADirectoryError: if the path given is actually a directory.
        """
        return self.write_problem(file_path, overwrite)

    def _write_to_stream(self, inp):
        """
        Writes the problem to a writeable stream.

        :param inp: Writable input file
        :type inp: MCNP_InputFile
        """
        with warnings.catch_warnings(record=True) as warning_catch:
            objects_list = []
            if self.message:
                objects_list.append(([self.message], False))
            objects_list += [
                ([self.title], False),
                (self.cells, True),
                (self.surfaces, True),
                (self.data_inputs, True),
            ]
            for objects, terminate in objects_list:
                for obj in objects:
                    lines = obj.format_for_mcnp_input(self.mcnp_version)
                    if warning_catch:
                        # handle ALL new warnings
                        for warning in warning_catch[::-1]:
                            if getattr(warning, "handled", None):
                                break
                            warning.lineno = inp.lineno
                            warning.path = inp.name
                            warning.obj = obj
                            warning.lines = lines
                            warning.handled = True
                    for line in lines:
                        inp.write(line + "\n")
                if terminate:
                    inp.write("\n")
            for line in self.cells._run_children_format_for_mcnp(
                self.data_inputs, self.mcnp_version
            ):
                inp.write(line + "\n")

            inp.write("\n")
        self._handle_warnings(warning_catch)

    def _handle_warnings(self, warning_queue):
        class WarningLevels(Enum):
            SUPRESS = 0
            MINIMAL = 1
            MAXIMAL = 5

        warning_level = WarningLevels.MAXIMAL

        for warning_message in warning_queue:
            warning = warning_message.message
            message = f"The input starting on Line {warning_message.lineno} of: {warning_message.path} expanded. "
            if warning_level == WarningLevels.SUPRESS:
                continue
            elif warning_level == WarningLevels.MINIMAL:
                if warning.cause == "value":
                    message += f"The new value is: {warning.new_value}"
                else:
                    message += f"The new lines are: {warning.new_value}"
            elif warning_level == WarningLevels.MAXIMAL:
                message += "\nThe new input is:\n"
                for i, line in enumerate(warning_message.lines):
                    message += f"     {warning_message.lineno + i:5g}| {line}\n"
                message += warning.message
            warning = LineExpansionWarning(message)
            warnings.warn(warning, stacklevel=3)

    def __load_data_inputs_to_object(self, data_inputs):
        """
        Loads data input into their appropriate problem attribute.

        Problem-level input should be loaded this way like: mode and kcode.
        """
        inputs_to_property = {mode.Mode: "_mode"}
        inputs_loaded = set()
        for input in data_inputs:
            if type(input) in inputs_to_property:
                if type(input) in inputs_loaded:
                    raise MalformedInputError(
                        input,
                        f"The input: {type(input)} is only allowed once in a problem",
                    )
                setattr(self, inputs_to_property[type(input)], input)
                inputs_loaded.add(type(input))

    def __str__(self):
        return f"MCNP problem for: {self._input_file}, {self._title}"

    def __repr__(self):
        ret = f"MCNP problem for: {self._input_file}\n"
        if self.message:
            ret += str(self._message) + "\n"
        ret += str(self._title) + "\n"
        for collection in [self.cells, self.surfaces, self.data_inputs]:
            for obj in collection:
                ret += f"{obj}\n"
            ret += "\n"
        return ret



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from .surface_type import SurfaceType
from .surface import Surface
from montepy.errors import *
from montepy.utilities import *


class AxisPlane(Surface):
    """
    Represents PX, PY, PZ

    :param input: The Input object representing the input
    :type input: Input
    """

    COORDINATE = {SurfaceType.PX: "x", SurfaceType.PY: "y", SurfaceType.PZ: "z"}

    def __init__(self, input=None):
        self._location = self._generate_default_node(float, None)
        super().__init__(input)
        ST = SurfaceType
        if input:
            if self.surface_type not in [ST.PX, ST.PY, ST.PZ]:
                raise ValueError("AxisPlane must be a surface of type: PX, PY, or PZ")
            if len(self.surface_constants) != 1:
                raise ValueError("AxisPlane must have exactly 1 surface constant")
            self._location = self._surface_constants[0]
        else:
            self._surface_constants = [self._location]

    @make_prop_val_node("_location", (float, int), float)
    def location(self):
        """
        The location of the plane in space.

        :rtype: float
        """
        pass

    def validate(self):
        super().validate()
        if self.location is None:
            raise IllegalState(f"Surface: {self.number} does not have a location set.")

    def find_duplicate_surfaces(self, surfaces, tolerance):
        ret = []
        # do not assume transform and periodic surfaces are the same.
        if not self.old_periodic_surface:
            for surface in surfaces:
                if surface != self and surface.surface_type == self.surface_type:
                    if not self.old_periodic_surface:
                        if abs(self.location - surface.location) < tolerance:
                            if self.transform:
                                if surface.transform:
                                    if self.transform.equivalent(
                                        surface.transform, tolerance
                                    ):
                                        ret.append(surface)
                            else:
                                if surface.transform is None:
                                    ret.append(surface)
            return ret
        else:
            return []



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from enum import Enum, unique


@unique
class BlockType(Enum):
    """
    An enumeration for the different blocks in an input file.
    """

    CELL = 0
    """
    The first block that details Cell information.
    """
    SURFACE = 1
    """
    The second block that details Surface information.
    """
    DATA = 2
    """
    The third block that provides additional information
    and data.
    """



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.input_parser.data_parser import DataParser
from montepy.input_parser import syntax_node


class TallyParser(DataParser):
    """
    A barebone parser for parsing tallies before they are fully implemented.

    .. versionadded:: 0.2.0

    :returns: a syntax tree for the data input.
    :rtype: SyntaxNode
    """

    debugfile = None

    @_("introduction tally_specification")
    def tally(self, p):
        ret = {}
        for key, node in p.introduction.nodes.items():
            ret[key] = node
        ret["data"] = p.tally_specification
        return syntax_node.SyntaxNode("data", ret)

    @_("tally_numbers", "tally_numbers end_phrase")
    def tally_specification(self, p):
        if hasattr(p, "end_phrase"):
            text = p.end_phrase
        else:
            text = syntax_node.ValueNode(None, str)

        return syntax_node.SyntaxNode(
            "tally list", {"tally": p.tally_numbers, "end": text}
        )

    @_("PARTICLE", "PARTICLE padding")
    def end_phrase(self, p):
        """
        A non-zero number with or without padding.

        :returns: a float ValueNode
        :rtype: ValueNode
        """
        return self._flush_phrase(p, str)

    @_(
        "tally_numbers tally_numbers",
        "number_sequence",
        "tally_group",
        "tally_numbers padding",
    )
    def tally_numbers(self, p):
        if hasattr(p, "tally_numbers"):
            ret = p.tally_numbers
            ret.nodes["right"] += p.padding
            return ret
        if hasattr(p, "tally_numbers1"):
            return syntax_node.SyntaxNode("tally tree", {"left": p[0], "right": p[1]})
        else:
            left = syntax_node.PaddingNode(None)
            right = syntax_node.PaddingNode(None)
            return syntax_node.SyntaxNode(
                "tally set", {"left": left, "tally": p[0], "right": right}
            )

    @_(
        '"(" number_sequence ")"',
        '"(" padding number_sequence ")"',
    )
    def tally_group(self, p):
        left = syntax_node.PaddingNode(p[0])
        if hasattr(p, "padding"):
            left.append(p.padding)
        right = syntax_node.PaddingNode(p[-1])
        return syntax_node.SyntaxNode(
            "tally set", {"left": left, "tally": p.number_sequence, "right": right}
        )



import gc
import montepy

import time
import tracemalloc

FAIL_THRESHOLD = 30

tracemalloc.start()
start = time.time()

problem = montepy.read_input("benchmark/big_model.imcnp")

stop = time.time()

print(f"Took {stop - start} seconds")
print(f"Memory usage report: {tracemalloc.get_traced_memory()[0]/1024/1024} MB")
del problem
gc.collect()
print(f"Memory usage report after GC: {tracemalloc.get_traced_memory()[0]/1024/1024} MB")

if (stop - start) > FAIL_THRESHOLD:
    raise RuntimeError(
        f"Benchmark took too long to complete. It must be faster than: {FAIL_THRESHOLD} s."
    )



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.input_parser.parser_base import MCNP_Parser
from montepy.input_parser import syntax_node


class ReadParser(MCNP_Parser):
    """
    A parser for handling "read" inputs.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.
    """

    debugfile = None
    dont_copy = {"parameter"}

    @_("KEYWORD padding parameters", "padding KEYWORD padding parameters")
    def read_input(self, p):
        if isinstance(p[0], syntax_node.PaddingNode):
            start_pad = p[0]
            mid_pad = p[2]
        else:
            start_pad = syntax_node.PaddingNode()
            mid_pad = p[1]
        if p.KEYWORD.lower() == "read":
            return syntax_node.SyntaxNode(
                "read",
                {
                    "start_pad": start_pad,
                    "keyword": p.KEYWORD,
                    "padding": mid_pad,
                    "parameters": p.parameters,
                },
            )
        else:
            return False

    @_(
        "classifier param_seperator file_phrase",
    )
    def parameter(self, p):
        return syntax_node.SyntaxNode(
            p.classifier.prefix.value,
            {"classifier": p.classifier, "seperator": p.param_seperator, "data": p[2]},
        )



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from abc import abstractmethod
import copy
import itertools
from montepy.errors import NumberConflictError
from montepy.mcnp_object import MCNP_Object
import montepy
from montepy.utilities import *


def _number_validator(self, number):
    if number < 0:
        raise ValueError("number must be >= 0")
    if self._problem:
        obj_map = montepy.MCNP_Problem._NUMBERED_OBJ_MAP
        try:
            collection_type = obj_map[type(self)]
        except KeyError as e:
            found = False
            for obj_class in obj_map:
                if isinstance(self, obj_class):
                    collection_type = obj_map[obj_class]
                    found = True
                    break
            if not found:
                raise e
        collection = getattr(self._problem, collection_type.__name__.lower())
        collection.check_number(number)
        collection._update_number(self.number, number, self)


class Numbered_MCNP_Object(MCNP_Object):

    @make_prop_val_node("_number", int, validator=_number_validator)
    def number(self):
        """
        The current number of the object that will be written out to a new input.

        :rtype: int
        """
        pass

    @property
    @abstractmethod
    def old_number(self):
        """
        The original number of the object provided in the input file

        :rtype: int
        """
        pass

    def clone(self, starting_number=None, step=None):
        """
        Create a new independent instance of this object with a new number.

        This relies mostly on ``copy.deepcopy``.

        .. note ::
            If starting_number, or step are not specified
            :func:`~montepy.numbered_object_collection.NumberedObjectCollection.starting_number`,
            and :func:`~montepy.numbered_object_collection.NumberedObjectCollection.step` are used as default values,
            if this object is tied to a problem.
            For instance a ``Material`` will use ``problem.materials`` default information.
            Otherwise ``1`` will be used as default values

        .. versionadded:: 0.5.0

        :param starting_number: The starting number to request for a new object number.
        :type starting_number: int
        :param step: the step size to use to find a new valid number.
        :type step: int
        :returns: a cloned copy of this object.
        :rtype: type(self)

        """
        if not isinstance(starting_number, (int, type(None))):
            raise TypeError(
                f"Starting_number must be an int. {type(starting_number)} given."
            )
        if not isinstance(step, (int, type(None))):
            raise TypeError(f"step must be an int. {type(step)} given.")
        if starting_number is not None and starting_number <= 0:
            raise ValueError(f"starting_number must be >= 1. {starting_number} given.")
        if step is not None and step <= 0:
            raise ValueError(f"step must be >= 1. {step} given.")
        ret = copy.deepcopy(self)
        if self._problem:
            ret.link_to_problem(self._problem)
            collection_type = montepy.MCNP_Problem._NUMBERED_OBJ_MAP[type(self)]
            collection = getattr(self._problem, collection_type.__name__.lower())
            if starting_number is None:
                starting_number = collection.starting_number
            if step is None:
                step = collection.step
            ret.number = collection.request_number(starting_number, step)
            collection.append(ret)
            return ret
        if starting_number is None:
            starting_number = 1
        if step is None:
            step = 1
        for number in itertools.count(starting_number, step):
            # only reached if not tied to a problem
            ret.number = number
            if number != self.number:
                return ret



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.surfaces.axis_plane import AxisPlane
from montepy.surfaces.surface import Surface
from montepy.surfaces.surface_type import SurfaceType
from montepy.surfaces.cylinder_on_axis import CylinderOnAxis
from montepy.surfaces.cylinder_par_axis import CylinderParAxis
from montepy.surfaces.general_plane import GeneralPlane


def surface_builder(input):
    """
    Builds a Surface object for the type of Surface

    .. versionchanged:: 0.2.0
        The ``comments`` argument has been removed with the simpler init function.

    :param input: The Input object representing the input
    :type input: Input
    :returns: A Surface object properly parsed. If supported a sub-class of Surface will be given.
    :rtype: Surface
    """
    ST = SurfaceType
    buffer_surface = Surface(input)
    type_of_surface = buffer_surface.surface_type
    if type_of_surface in [ST.C_X, ST.C_Y, ST.C_Z]:
        return CylinderParAxis(input)
    elif type_of_surface in [ST.CX, ST.CY, ST.CZ]:
        return CylinderOnAxis(input)
    elif type_of_surface in [ST.PX, ST.PY, ST.PZ]:
        return AxisPlane(input)
    elif type_of_surface == ST.P:
        return GeneralPlane(input)
    else:
        return buffer_surface



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.input_parser.data_parser import DataParser
from montepy.input_parser import syntax_node


class ThermalParser(DataParser):
    """
    A parser for thermal scattering law inputs.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    :rtype: SyntaxNode
    """

    debugfile = None

    @_("introduction thermal_law_sequence")
    def thermal_mat(self, p):
        ret = {}
        for key, node in p.introduction.nodes.items():
            ret[key] = node
        ret["data"] = p.thermal_law_sequence
        return syntax_node.SyntaxNode("thermal scattering", ret)

    @_("thermal_law", "thermal_law_sequence thermal_law")
    def thermal_law_sequence(self, p):
        if len(p) == 1:
            sequence = syntax_node.ListNode("thermal law sequence")
            sequence.append(p[0])
        else:
            sequence = p[0]
            sequence.append(p[1])
        return sequence

    @_("THERMAL_LAW", "THERMAL_LAW padding")
    def thermal_law(self, p):
        return self._flush_phrase(p, str)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from . import axis_plane
from .axis_plane import AxisPlane
from . import cylinder_par_axis
from .cylinder_par_axis import CylinderParAxis
from . import cylinder_on_axis
from .cylinder_on_axis import CylinderOnAxis
from . import half_space
from . import surface
from . import surface_builder



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import hypothesis
from hypothesis import given, settings, strategies as st
import copy
import montepy
import montepy.cells
from montepy.errors import NumberConflictError
import unittest
import pytest
import os


class TestNumberedObjectCollection(unittest.TestCase):
    def setUp(self):
        self.simple_problem = montepy.read_input("tests/inputs/test.imcnp")

    def test_bad_init(self):
        with self.assertRaises(TypeError):
            montepy.cells.Cells(5)

    def test_numbers(self):
        cell_numbers = [1, 2, 3, 99, 5]
        surf_numbers = [1000, 1005, 1010, 1015, 1020, 1025]
        mat_numbers = [1, 2, 3]
        problem = self.simple_problem
        self.assertEqual(list(problem.cells.numbers), cell_numbers)
        self.assertEqual(list(problem.surfaces.numbers), surf_numbers)
        self.assertEqual(list(problem.materials.numbers), mat_numbers)

    def test_number_conflict_init(self):
        cells = list(self.simple_problem.cells)
        cells.append(cells[1])
        with self.assertRaises(NumberConflictError):
            montepy.cells.Cells(cells)

    def test_check_number(self):
        with self.assertRaises(NumberConflictError):
            self.simple_problem.cells.check_number(1)
        with self.assertRaises(TypeError):
            self.simple_problem.cells.check_number("5")
        # testing a number that shouldn't conflict to ensure error isn't raised
        self.simple_problem.cells.check_number(20)

    def test_objects(self):
        generated = list(self.simple_problem.cells)
        objects = self.simple_problem.cells.objects
        self.assertEqual(generated, objects)

    def test_pop(self):
        cells = copy.deepcopy(self.simple_problem.cells)
        size = len(cells)
        target = list(cells)[-1]
        popped = cells.pop()
        self.assertEqual(target, popped)
        self.assertEqual(size - 1, len(cells))
        with self.assertRaises(TypeError):
            cells.pop("hi")

    def test_extend(self):
        surfaces = copy.deepcopy(self.simple_problem.surfaces)
        extender = list(surfaces)[0:2]
        size = len(surfaces)
        with self.assertRaises(NumberConflictError):
            surfaces.extend(extender)
        self.assertEqual(len(surfaces), size)
        extender = copy.deepcopy(extender)
        extender[0].number = 50
        extender[1].number = 60
        surfaces.extend(extender)
        self.assertEqual(len(surfaces), size + 2)
        # force a num_cache miss
        extender = copy.deepcopy(extender)
        for surf in extender:
            surf._problem = None
        surfaces[1000].number = 1
        extender[0].number = 1000
        extender[1].number = 70
        surfaces.extend(extender)
        self.assertEqual(len(surfaces), size + 4)
        with self.assertRaises(TypeError):
            surfaces.extend(5)
        with self.assertRaises(TypeError):
            surfaces.extend([5])

    def test_iter(self):
        size = len(self.simple_problem.cells)
        counter = 0
        for cell in self.simple_problem.cells:
            counter += 1
        self.assertEqual(size, counter)

    def test_append(self):
        cells = copy.deepcopy(self.simple_problem.cells)
        cell = copy.deepcopy(cells[1])
        size = len(cells)
        with self.assertRaises(NumberConflictError):
            cells.append(cell)
        with self.assertRaises(TypeError):
            cells.append(5)
        cell.number = 20
        cells.append(cell)
        self.assertEqual(len(cells), size + 1)

    def test_append_renumber(self):
        cells = copy.deepcopy(self.simple_problem.cells)
        size = len(cells)
        cell = copy.deepcopy(cells[1])
        cell.number = 20
        cells.append_renumber(cell)
        self.assertEqual(len(cells), size + 1)
        with self.assertRaises(TypeError):
            cells.append_renumber(5)
        with self.assertRaises(TypeError):
            cells.append_renumber(cell, "hi")
        cell = copy.deepcopy(cell)
        cell._problem = None
        cell.number = 1
        cells.append_renumber(cell)
        self.assertEqual(cell.number, 4)
        self.assertEqual(len(cells), size + 2)

    def test_append_renumber_problems(self):
        print(hex(id(self.simple_problem.materials._problem)))
        prob1 = copy.deepcopy(self.simple_problem)
        prob2 = copy.deepcopy(self.simple_problem)
        print(hex(id(self.simple_problem.materials._problem)))
        # Delete Material 2, making its number available.
        prob2.materials.remove(prob2.materials[2])
        len_mats = len(prob2.materials)
        mat1 = prob1.materials[1]
        new_num = prob2.materials.append_renumber(mat1)
        assert new_num == 2, "Material not renumbered correctly."
        assert len(prob2.materials) == len_mats + 1, "Material not appended"
        assert prob2.materials[2] is mat1, "Material 2 is not the new material"

    def test_request_number(self):
        cells = self.simple_problem.cells
        self.assertEqual(cells.request_number(6), 6)
        self.assertEqual(cells.request_number(1), 4)
        self.assertEqual(cells.request_number(99, 6), 105)
        with self.assertRaises(TypeError):
            cells.request_number("5")
        with self.assertRaises(TypeError):
            cells.request_number(1, "5")

    def test_next_number(self):
        cells = self.simple_problem.cells
        self.assertEqual(cells.next_number(), 100)
        self.assertEqual(cells.next_number(6), 105)
        with self.assertRaises(TypeError):
            cells.next_number("5")
        with self.assertRaises(ValueError):
            cells.next_number(-1)

    def test_getitem(self):
        cells = self.simple_problem.cells
        list_version = list(cells)
        self.assertEqual(cells[1], list_version[0])
        # force stale cache misses
        cells[1].number = 20
        with self.assertRaises(KeyError):
            cells[1]
        # force cache miss
        self.assertEqual(cells[20], list_version[0])
        with self.assertRaises(TypeError):
            cells["5"]

    def test_delete(self):
        cells = copy.deepcopy(self.simple_problem.cells)
        size = len(cells)
        del cells[1]
        self.assertEqual(size - 1, len(cells))
        with self.assertRaises(TypeError):
            del cells["5"]

    def test_setitem(self):
        cells = copy.deepcopy(self.simple_problem.cells)
        cell = cells[1]
        size = len(cells)
        with self.assertRaises(NumberConflictError):
            cells[1] = cell
        with self.assertRaises(TypeError):
            cells[1] = 5
        with self.assertRaises(TypeError):
            cells["1"] = cell
        cell = copy.deepcopy(cell)
        cell.number = 20
        cells[50] = cell
        self.assertEqual(len(cells), size + 1)

    def test_iadd(self):
        cells = copy.deepcopy(self.simple_problem.cells)
        list_cells = list(cells)
        size = len(cells)
        with self.assertRaises(NumberConflictError):
            cells += list_cells
        with self.assertRaises(NumberConflictError):
            cells += montepy.cells.Cells(list_cells)

        with self.assertRaises(TypeError):
            cells += 5
        with self.assertRaises(TypeError):
            cells += [5]

        list_cells = [copy.deepcopy(cells[1])]
        list_cells[0].number = 20
        cells += list_cells
        self.assertEqual(len(cells), size + 1)

        this_problem = copy.deepcopy(self.simple_problem)
        for cell in this_problem.cells:
            cell.number += 1000
        this_problem.cells += self.simple_problem.cells
        self.assertEqual(len(this_problem.cells), size * 2)

    def test_slice(self):
        test_numbers = [c.number for c in self.simple_problem.cells[1:5]]
        self.assertEqual([1, 2, 3, 5], test_numbers)
        test_numbers = [c.number for c in self.simple_problem.cells[2:]]
        self.assertEqual([2, 3, 5, 99], test_numbers)
        test_numbers = [c.number for c in self.simple_problem.cells[::-3]]
        self.assertEqual([99, 3], test_numbers)
        test_numbers = [c.number for c in self.simple_problem.cells[:6:3]]
        self.assertEqual([3], test_numbers)
        test_numbers = [c.number for c in self.simple_problem.cells[5::-1]]
        self.assertEqual([5, 3, 2, 1], test_numbers)
        test_numbers = [s.number for s in self.simple_problem.surfaces[1000::10]]
        self.assertEqual([1000, 1010, 1020], test_numbers)
        test_numbers = [s.number for s in self.simple_problem.surfaces[:]]
        self.assertEqual([1000, 1005, 1010, 1015, 1020, 1025], test_numbers)
        test_numbers = [m.number for m in self.simple_problem.materials[:2]]
        self.assertEqual([1, 2], test_numbers)
        test_numbers = [m.number for m in self.simple_problem.materials[::2]]
        self.assertEqual([2], test_numbers)

    def test_get(self):
        cell_found = self.simple_problem.cells.get(1)
        self.assertEqual(self.simple_problem.cells[1], cell_found)
        surf_not_found = self.simple_problem.surfaces.get(39)  # 39 buried, 0 found
        self.assertIsNone(surf_not_found)
        default_mat = self.simple_problem.materials[3]
        self.assertEqual(
            self.simple_problem.materials.get(42, default_mat), default_mat
        )

    def test_keys(self):
        cell_nums = []
        for c in self.simple_problem.cells:
            cell_nums.append(c.number)
        cell_keys = []
        for k in self.simple_problem.cells.keys():
            cell_keys.append(k)
        self.assertEqual(cell_nums, cell_keys)

    def test_values(self):
        list_cells = list(self.simple_problem.cells)
        list_values = list(self.simple_problem.cells.values())
        self.assertEqual(list_cells, list_values)

    def test_items(self):
        zipped = zip(
            self.simple_problem.cells.keys(), self.simple_problem.cells.values()
        )
        cell_items = self.simple_problem.cells.items()
        self.assertTupleEqual(tuple(zipped), tuple(cell_items))

    def test_surface_generators(self):
        answer_num = [1000, 1010]
        spheres = list(self.simple_problem.surfaces.so)
        self.assertEqual(len(answer_num), len(spheres))
        for i, sphere in enumerate(spheres):
            self.assertEqual(answer_num[i], sphere.number)

    def test_number_adding_concurancy(self):
        surfaces = copy.deepcopy(self.simple_problem.surfaces)
        new_surf = copy.deepcopy(surfaces[1005])
        new_surf.number = 5
        surfaces.append(new_surf)
        size = len(surfaces)
        new_surf = copy.deepcopy(new_surf)
        with self.assertRaises(NumberConflictError):
            surfaces.append(new_surf)
        surfaces.append_renumber(new_surf)
        self.assertEqual(len(surfaces), size + 1)
        self.assertEqual(new_surf.number, 6)

    def test_str(self):
        cells = self.simple_problem.cells
        self.assertEqual(str(cells), "Cells: [1, 2, 3, 99, 5]")
        key_phrases = [
            "Numbered_object_collection: obj_class: <class 'montepy.cell.Cell'>",
            "Objects: [CELL: 1",
            "Number cache: {1: CELL: 1",
        ]
        for phrase in key_phrases:
            self.assertIn(phrase, repr(cells))


# test data numbered object
@pytest.fixture(scope="module")
def read_simple_problem():
    return montepy.read_input(os.path.join("tests", "inputs", "test.imcnp"))


@pytest.fixture
def cp_simple_problem(read_simple_problem):
    return copy.deepcopy(read_simple_problem)


def test_data_init(cp_simple_problem):
    new_mats = montepy.materials.Materials(
        list(cp_simple_problem.materials), problem=cp_simple_problem
    )
    assert list(new_mats) == list(cp_simple_problem.materials)


def test_data_append(cp_simple_problem):
    prob = cp_simple_problem
    new_mat = copy.deepcopy(next(iter(prob.materials)))
    new_mat.number = prob.materials.request_number()
    prob.materials.append(new_mat)
    assert new_mat in prob.materials
    assert new_mat in prob.data_inputs
    assert prob.data_inputs.count(new_mat) == 1
    # trigger getting data_inputs end
    prob.materials.clear()
    prob.materials.append(new_mat)
    assert new_mat in prob.materials
    assert new_mat in prob.data_inputs
    assert prob.data_inputs.count(new_mat) == 1
    prob.data_inputs.clear()
    prob.materials._last_index = None
    new_mat = copy.deepcopy(next(iter(prob.materials)))
    new_mat.number = prob.materials.request_number()
    prob.materials.append(new_mat)
    assert new_mat in prob.materials
    assert new_mat in prob.data_inputs
    assert prob.data_inputs.count(new_mat) == 1
    # trigger getting index of last material
    prob.materials._last_index = None
    new_mat = copy.deepcopy(next(iter(prob.materials)))
    new_mat.number = prob.materials.request_number()
    prob.materials.append(new_mat)
    assert new_mat in prob.materials
    assert new_mat in prob.data_inputs
    assert prob.data_inputs.count(new_mat) == 1


def test_data_append_renumber(cp_simple_problem):
    prob = cp_simple_problem
    new_mat = copy.deepcopy(next(iter(prob.materials)))
    prob.materials.append_renumber(new_mat)
    assert new_mat in prob.materials
    assert new_mat in prob.data_inputs
    assert prob.data_inputs.count(new_mat) == 1


def test_data_remove(cp_simple_problem):
    prob = cp_simple_problem
    old_mat = next(iter(prob.materials))
    prob.materials.remove(old_mat)
    assert old_mat not in prob.materials
    assert old_mat not in prob.data_inputs


def test_data_delete(cp_simple_problem):
    prob = cp_simple_problem
    old_mat = next(iter(prob.materials))
    del prob.materials[old_mat.number]
    assert old_mat not in prob.materials
    assert old_mat not in prob.data_inputs
    with pytest.raises(TypeError):
        del prob.materials["foo"]


def test_data_clear(cp_simple_problem):
    data_len = len(cp_simple_problem.data_inputs)
    mat_len = len(cp_simple_problem.materials)
    cp_simple_problem.materials.clear()
    assert len(cp_simple_problem.materials) == 0
    assert len(cp_simple_problem.data_inputs) == data_len - mat_len


def test_data_pop(cp_simple_problem):
    old_mat = next(reversed(list(cp_simple_problem.materials)))
    popper = cp_simple_problem.materials.pop()
    assert popper is old_mat
    assert old_mat not in cp_simple_problem.materials
    assert old_mat not in cp_simple_problem.data_inputs
    with pytest.raises(TypeError):
        cp_simple_problem.materials.pop("foo")


# disable function scoped fixtures
@settings(suppress_health_check=[hypothesis.HealthCheck.function_scoped_fixture])
@given(start_num=st.integers(), step=st.integers())
def test_num_collect_clone(cp_simple_problem, start_num, step):
    surfs = copy.deepcopy(cp_simple_problem.surfaces)
    if start_num <= 0 or step <= 0:
        with pytest.raises(ValueError):
            surfs.clone(start_num, step)
        return
    for clear in [False, True]:
        if clear:
            surfs.link_to_problem(None)
        new_surfs = surfs.clone(start_num, step)
        for new_surf, old_surf in zip(new_surfs, surfs):
            assert new_surf is not old_surf
            assert new_surf.surface_type == old_surf.surface_type
            assert new_surf.number != old_surf.number


def test_num_collect_clone_default(cp_simple_problem):
    surfs = copy.deepcopy(cp_simple_problem.surfaces)
    for clear in [False, True]:
        if clear:
            surfs.link_to_problem(None)
        new_surfs = surfs.clone()
        for new_surf, old_surf in zip(new_surfs, surfs):
            assert new_surf is not old_surf
            assert new_surf.surface_type == old_surf.surface_type
            assert new_surf.number != old_surf.number


@pytest.mark.parametrize(
    "args, error",
    [
        (("c", 1), TypeError),
        ((1, "d"), TypeError),
        ((-1, 1), ValueError),
        ((0, 1), ValueError),
        ((1, 0), ValueError),
        ((1, -1), ValueError),
    ],
)
def test_num_collect_clone_bad(cp_simple_problem, args, error):
    surfs = cp_simple_problem.surfaces
    with pytest.raises(error):
        surfs.clone(*args)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.constants import BLANK_SPACE_CONTINUE
import functools
import re

"""
A package for helper universal utility functions
"""


def fortran_float(number_string):
    """
    Attempts to convert a FORTRAN formatted float string to a float.

    FORTRAN allows silly things for scientific notation like ``6.02+23``
    to represent Avogadro's Number.

    :param number_string: the string that will be converted to a float
    :type number_string: str
    :raises ValueError: If the string can not be parsed as a float.
    :return: the parsed float of the this string
    :rtype: float
    """
    try:
        return float(number_string)

    except ValueError as e:
        update_number = re.sub(r"(\d)([-+])", r"\1E\2", number_string)
        try:
            return float(update_number)
        except ValueError:
            raise ValueError(f"Value Not parsable as float: {number_string}") from e


def is_comment(line):
    """
    Determines if the line is a ``C comment`` style comment.

    :param line: the line to analyze
    :type line: str
    :returns: True if the line is a comment
    :rtype: bool
    """
    upper_start = line[0 : BLANK_SPACE_CONTINUE + 1].upper()
    non_blank_comment = upper_start and line.lstrip().upper().startswith("C ")
    if non_blank_comment:
        return True
    blank_comment = ("C" == upper_start.strip() and "\n" in line) or (
        "C" == upper_start and "\n" not in line
    )
    return blank_comment


def make_prop_val_node(
    hidden_param, types=None, base_type=None, validator=None, deletable=False
):
    """
    A decorator function for making a property from a ValueNode.

    This decorator is meant to handle all boiler plate. It will get and
    set the value property of the underlying ValueNode.
    By default the property is not settable unless types is set.

    :param hidden_param: The string representing the parameter name of the internally stored ValueNode.
    :type hidden_param: str
    :param types: the acceptable types for the settable, which is passed to isinstance. If an empty tuple will be
                type(self).
    :type types: Class, tuple
    :param validator: A validator function to run on values before setting. Must accept func(self, value).
    :type validator: function
    :param deletable: If true make this property deletable. When deleted the value will be set to None.
    :type deletable: bool
    """

    def decorator(func):
        @property
        @functools.wraps(func)
        def getter(self):
            result = func(self)
            if result:
                return result
            else:
                val = getattr(self, hidden_param)
                if val is None:
                    return None
                return val.value

        if types is not None:

            def setter(self, value):
                nonlocal types
                if isinstance(types, tuple) and len(types) == 0:
                    types = type(self)
                if not isinstance(value, types):
                    raise TypeError(
                        f"{func.__name__} must be of type: {types}. {value} given."
                    )
                if (
                    base_type is not None
                    and value is not None
                    and not isinstance(value, base_type)
                ):
                    value = base_type(value)
                if validator:
                    validator(self, value)
                node = getattr(self, hidden_param)
                node.value = value

            getter = getter.setter(setter)

        if deletable:

            def deleter(self):
                setattr(self, hidden_param, None)

            getter = getter.deleter(deleter)
        return getter

    return decorator


def make_prop_pointer(
    hidden_param, types=None, base_type=None, validator=None, deletable=False
):
    """
    A decorator function that makes a property based off of a pointer to another object.

    Note this can also be used for almost any circumstance as everything in python is a pointer.

    :param hidden_param: The string representing the parameter name of the internally stored ValueNode.
    :type hidden_param: str
    :param types: the acceptable types for the settable, which is passed to isinstance, if an empty tuple is provided the type will be self.
    :type types: Class, tuple
    :param validator: A validator function to run on values before setting. Must accept func(self, value).
    :type validator: function
    :param deletable: If true make this property deletable. When deleted the value will be set to None.
    :type deletable: bool
    """

    def decorator(func):
        @property
        @functools.wraps(func)
        def getter(self):
            result = func(self)
            if result:
                return result
            return getattr(self, hidden_param)

        if types is not None:

            def setter(self, value):
                nonlocal types
                if isinstance(types, tuple) and len(types) == 0:
                    types = type(self)
                if not isinstance(value, types):
                    raise TypeError(f"{func.__name__} must be of type: {types}")
                if base_type is not None and not isinstance(value, base_type):
                    value = base_type(value)
                if validator:
                    validator(self, value)
                setattr(self, hidden_param, value)

            getter = getter.setter(setter)
        if deletable:

            def deleter(self):
                setattr(self, hidden_param, None)

            getter = getter.deleter(deleter)
        return getter

    return decorator



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from .surface_type import SurfaceType
from .surface import Surface
from montepy.errors import *
from montepy.utilities import *


def _enforce_positive_radius(self, value):
    if value < 0.0:
        raise ValueError(f"Radius must be positive. {value} given")


class CylinderParAxis(Surface):
    """
    Represents surfaces: C/X, C/Y, C/Z

    :param input: The Input object representing the input
    :type input: Input
    """

    COORDINATE_PAIRS = {
        SurfaceType.C_X: {0: "y", 1: "z"},
        SurfaceType.C_Y: {0: "x", 1: "z"},
        SurfaceType.C_Z: {0: "x", 1: "y"},
    }
    """Which coordinate is what value for each cylinder type.
    """

    def __init__(self, input=None):
        self._coordinates = [
            self._generate_default_node(float, None),
            self._generate_default_node(float, None),
        ]
        self._radius = self._generate_default_node(float, None)
        super().__init__(input)
        ST = SurfaceType
        if input:
            if self.surface_type not in [ST.C_X, ST.C_Y, ST.C_Z]:
                raise ValueError(
                    "CylinderParAxis must be a surface of types: C/X, C/Y, C/Z"
                )
            if len(self.surface_constants) != 3:
                raise ValueError(
                    "CylinderParAxis must have exactly 3 surface_constants"
                )
            self._coordinates = self._surface_constants[0:2]
            self._radius = self._surface_constants[2]
        else:
            self._surface_constants = [*self._coordinates, self._radius]

    @property
    def coordinates(self):
        """
        The two coordinates for this cylinder to center on.

        :rytpe: tuple
        """
        return (self._coordinates[0].value, self._coordinates[1].value)

    @coordinates.setter
    def coordinates(self, coordinates):
        if not isinstance(coordinates, (list, tuple)):
            raise TypeError("coordinates must be a list")
        if len(coordinates) != 2:
            raise ValueError("coordinates must have exactly two elements")
        for val in coordinates:
            if not isinstance(val, (float, int)):
                raise TypeError(f"Coordinate must be a number. {val} given.")
        for i, val in enumerate(coordinates):
            self._coordinates[i].value = val

    @make_prop_val_node(
        "_radius", (float, int), float, validator=_enforce_positive_radius
    )
    def radius(self):
        """
        The radius of the cylinder.

        :rtype: float
        """
        pass

    def validate(self):
        super().validate()
        if self.radius is None:
            raise IllegalState(f"Surface: {self.number} does not have a radius set.")
        if any({c is None for c in self.coordinates}):
            raise IllegalState(f"Surface: {self.number} does not have coordinates set.")

    def find_duplicate_surfaces(self, surfaces, tolerance):
        ret = []
        # do not assume transform and periodic surfaces are the same.
        if not self.old_periodic_surface:
            for surface in surfaces:
                if surface != self and surface.surface_type == self.surface_type:
                    if not self.old_periodic_surface:
                        match = True
                        if abs(self.radius - surface.radius) >= tolerance:
                            match = False
                        for i, coordinate in enumerate(self.coordinates):
                            if abs(coordinate - surface.coordinates[i]) >= tolerance:
                                match = False
                        if match:
                            if self.transform:
                                if surface.transform:
                                    if self.transform.equivalent(
                                        surface.transform, tolerance
                                    ):
                                        ret.append(surface)
                            else:
                                if surface.transform is None:
                                    ret.append(surface)
            return ret
        else:
            return []



import cProfile
import montepy
import pstats

stats = cProfile.run(
    'montepy.read_input("benchmark/big_model.imcnp")',
    "prof/big.prof",
    sort=pstats.SortKey.CUMULATIVE,
)

stats = pstats.Stats("prof/big.prof")
stats.sort_stats(pstats.SortKey.CUMULATIVE, pstats.SortKey.TIME).print_stats(
    "montepy", 70
)
stats.sort_stats(pstats.SortKey.CUMULATIVE, pstats.SortKey.TIME).print_stats("sly", 20)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from abc import ABC, abstractmethod
import typing
import weakref

import montepy
from montepy.numbered_mcnp_object import Numbered_MCNP_Object
from montepy.errors import *
from montepy.utilities import *


def _enforce_positive(self, num):
    if num <= 0:
        raise ValueError(f"Value must be greater than 0. {num} given.")


class NumberedObjectCollection(ABC):
    """A collections of MCNP objects.

    It quacks like a dict, it acts like a dict, but it's a list.

    The items in the collection are accessible by their number.
    For instance to get the Cell with a number of 2 you can just say:

    ``problem.cells[2]``

    You can also add delete items like you would in a dictionary normally.

    Unlike dictionaries this collection also supports slices e.g., ``[1:3]``.
    This will return a new :class:`NumberedObjectCollection` with objects
    that have cell numbers that fit that slice. If a number is in a slice that
    is not an actual object it will just be skipped.

    Because MCNP numbered objects start at 1, so do the indices.
    The slices are effectively 1-based and endpoint-inclusive.
    This means rather than the normal behavior of [0:5] excluding the index
    5, 5 would be included.

    :param obj_class: the class of numbered objects being collected
    :type obj_class: type
    :param objects: the list of cells to start with if needed
    :type objects: list
    :param problem: the problem to link this collection to.
    :type problem: MCNP_Problem
    """

    def __init__(self, obj_class, objects=None, problem=None):
        self.__num_cache = {}
        assert issubclass(obj_class, Numbered_MCNP_Object)
        self._obj_class = obj_class
        self._objects = []
        self._start_num = 1
        self._step = 1
        self._problem_ref = None
        if problem is not None:
            self._problem_ref = weakref.ref(problem)
        if objects:
            if not isinstance(objects, list):
                raise TypeError("NumberedObjectCollection must be built from a list")
            for obj in objects:
                if not isinstance(obj, obj_class):
                    raise TypeError(
                        f"The object: {obj} being added to a NumberedObjectCollection is not of type {obj_class}"
                    )
                if obj.number in self.__num_cache:
                    raise NumberConflictError(
                        (
                            f"When building {self} there was a numbering conflict between: "
                            f"{obj} and {self[obj.number]}"
                        )
                    )
                self.__num_cache[obj.number] = obj
            self._objects = objects

    def link_to_problem(self, problem):
        """Links the card to the parent problem for this card.

        This is done so that cards can find links to other objects.

        :param problem: The problem to link this card to.
        :type problem: MCNP_Problem
        """
        if not isinstance(problem, (montepy.mcnp_problem.MCNP_Problem, type(None))):
            raise TypeError("problem must be an MCNP_Problem")
        if problem is None:
            self._problem_ref = None
        else:
            self._problem_ref = weakref.ref(problem)
        for obj in self:
            obj.link_to_problem(problem)

    @property
    def _problem(self):
        if self._problem_ref is not None:
            return self._problem_ref()
        return None

    def __getstate__(self):
        state = self.__dict__.copy()
        weakref_key = "_problem_ref"
        if weakref_key in state:
            del state[weakref_key]
        return state

    def __setstate__(self, crunchy_data):
        crunchy_data["_problem_ref"] = None
        self.__dict__.update(crunchy_data)

    @property
    def numbers(self):
        """
        A generator of the numbers being used.

        :rtype: generator
        """
        for obj in self._objects:
            # update cache every time we go through all objects
            self.__num_cache[obj.number] = obj
            yield obj.number

    def check_number(self, number):
        """Checks if the number is already in use, and if so raises an error.

        :param number: The number to check.
        :type number: int
        :raises NumberConflictError: if this number is in use.
        """
        if not isinstance(number, int):
            raise TypeError("The number must be an int")
        conflict = False
        # only can trust cache if being
        if self._problem:
            if number in self.__num_cache:
                conflict = True
        else:
            if number in self.numbers:
                conflict = True
        if conflict:
            raise NumberConflictError(
                f"Number {number} is already in use for the collection: {type(self)} by {self[number]}"
            )

    def _update_number(self, old_num, new_num, obj):
        """
        Updates the number associated with a specific object in the internal cache.

        :param old_num: the previous number the object had.
        :type old_num: int
        :param new_num: the number that is being set to.
        :type new_num: int
        :param obj: the object being updated.
        :type obj: self._obj_class
        """
        # don't update numbers you don't own
        if self.__num_cache.get(old_num, None) is not obj:
            return
        self.__num_cache.pop(old_num, None)
        self.__num_cache[new_num] = obj

    @property
    def objects(self):
        """
        Returns a shallow copy of the internal objects list.

        The list object is a new instance, but the underlying objects
        are the same.

        :rtype: list
        """
        return self._objects[:]

    def pop(self, pos=-1):
        """
        Pop the final items off of the collection

        :param pos: The index of the element to pop from the internal list.
        :type pos: int
        :return: the final elements
        :rtype: Numbered_MCNP_Object
        """
        if not isinstance(pos, int):
            raise TypeError("The index for popping must be an int")
        obj = self._objects.pop(pos)
        self.__num_cache.pop(obj.number, None)
        return obj

    def clear(self):
        """
        Removes all objects from this collection.
        """
        self._objects.clear()
        self.__num_cache.clear()

    def extend(self, other_list):
        """
        Extends this collection with another list.

        :param other_list: the list of objects to add.
        :type other_list: list
        :raises NumberConflictError: if these items conflict with existing elements.
        """
        if not isinstance(other_list, (list, type(self))):
            raise TypeError("The extending list must be a list")
        if self._problem:
            nums = set(self.__num_cache)
        else:
            nums = set(self.numbers)
        for obj in other_list:
            if not isinstance(obj, self._obj_class):
                raise TypeError(
                    "The object in the list {obj} is not of type: {self._obj_class}"
                )
            if obj.number in nums:
                raise NumberConflictError(
                    (
                        f"When adding to {type(self)} there was a number collision due to "
                        f"adding {obj} which conflicts with {self[obj.number]}"
                    )
                )
            nums.add(obj.number)
        self._objects.extend(other_list)
        self.__num_cache.update({obj.number: obj for obj in other_list})
        if self._problem:
            for obj in other_list:
                obj.link_to_problem(self._problem)

    def remove(self, delete):
        """
        Removes the given object from the collection.

        :param delete: the object to delete
        :type delete: Numbered_MCNP_Object
        """
        self.__num_cache.pop(delete.number, None)
        self._objects.remove(delete)

    def clone(self, starting_number=None, step=None):
        """
        Create a new instance of this collection, with all new independent
        objects with new numbers.

        This relies mostly on ``copy.deepcopy``.

        .. note ::
            If starting_number, or step are not specified :func:`starting_number`,
            and :func:`step` are used as default values.

        .. versionadded:: 0.5.0

        :param starting_number: The starting number to request for a new object numbers.
        :type starting_number: int
        :param step: the step size to use to find a new valid number.
        :type step: int
        :returns: a cloned copy of this object.
        :rtype: type(self)

        """
        if not isinstance(starting_number, (int, type(None))):
            raise TypeError(
                f"Starting_number must be an int. {type(starting_number)} given."
            )
        if not isinstance(step, (int, type(None))):
            raise TypeError(f"step must be an int. {type(step)} given.")
        if starting_number is not None and starting_number <= 0:
            raise ValueError(f"starting_number must be >= 1. {starting_number} given.")
        if step is not None and step <= 0:
            raise ValueError(f"step must be >= 1. {step} given.")
        if starting_number is None:
            starting_number = self.starting_number
        if step is None:
            step = self.step
        objs = []
        for obj in self:
            new_obj = obj.clone(starting_number, step)
            starting_number = new_obj.number
            objs.append(new_obj)
            starting_number = new_obj.number + step
        return type(self)(objs)

    @make_prop_pointer("_start_num", int, validator=_enforce_positive)
    def starting_number(self):
        """
        The starting number to use when an object is cloned.

        :returns: the starting number
        :rtype: int
        """
        pass

    @make_prop_pointer("_step", int, validator=_enforce_positive)
    def step(self):
        """
        The step size to use to find a valid number during cloning.

        :returns: the step size
        :rtype: int
        """
        pass

    def __iter__(self):
        self._iter = self._objects.__iter__()
        return self._iter

    def __str__(self):
        base_class_name = self.__class__.__name__
        numbers = list(self.numbers)
        return f"{base_class_name}: {numbers}"

    def __repr__(self):
        return (
            f"Numbered_object_collection: obj_class: {self._obj_class}, problem: {self._problem}\n"
            f"Objects: {self._objects}\n"
            f"Number cache: {self.__num_cache}"
        )

    def append(self, obj):
        """Appends the given object to the end of this collection.

        :param obj: the object to add.
        :type obj: Numbered_MCNP_Object
        :raises NumberConflictError: if this object has a number that is already in use.
        """
        if not isinstance(obj, self._obj_class):
            raise TypeError(f"object being appended must be of type: {self._obj_class}")
        self.check_number(obj.number)
        self.__num_cache[obj.number] = obj
        self._objects.append(obj)
        if self._problem:
            obj.link_to_problem(self._problem)

    def append_renumber(self, obj, step=1):
        """Appends the object, but will renumber the object if collision occurs.

        This behaves like append, except if there is a number collision the object will
        be renumbered to an available number. The number will be incremented by step
        until an available number is found.

        :param obj: The MCNP object being added to the collection.
        :type obj: Numbered_MCNP_Object
        :param step: the incrementing step to use to find a new number.
        :type step: int
        :return: the number for the object.
        :rtype: int
        """
        if not isinstance(obj, self._obj_class):
            raise TypeError(f"object being appended must be of type: {self._obj_class}")
        if not isinstance(step, int):
            raise TypeError("The step number must be an int")
        number = obj.number
        if self._problem:
            obj.link_to_problem(self._problem)
        try:
            self.append(obj)
        except NumberConflictError:
            number = self.request_number(number, step)
            obj.number = number
            self.append(obj)

        return number

    def request_number(self, start_num=None, step=None):
        """Requests a new available number.

        This method does not "reserve" this number. Objects
        should be immediately added to avoid possible collisions
        caused by shifting numbers of other objects in the collection.

        .. note ::
            If starting_number, or step are not specified :func:`starting_number`,
            and :func:`step` are used as default values.

        .. versionchanged:: 0.5.0
            In 0.5.0 the default values were changed to reference :func:`starting_number` and :func:`step`.

        :param start_num: the starting number to check.
        :type start_num: int
        :param step: the increment to jump by to find new numbers.
        :type step: int
        :returns: an available number
        :rtype: int
        """
        if not isinstance(start_num, (int, type(None))):
            raise TypeError("start_num must be an int")
        if not isinstance(step, (int, type(None))):
            raise TypeError("step must be an int")
        if start_num is None:
            start_num = self.starting_number
        if step is None:
            step = self.step
        number = start_num
        while True:
            try:
                self.check_number(number)
                break
            except NumberConflictError:
                number += step
        return number

    def next_number(self, step=1):
        """Get the next available number, based on the maximum number.

        This works by finding the current maximum number, and then adding the
        stepsize to it.

        :param step: how much to increase the last number by
        :type step: int
        """
        if not isinstance(step, int):
            raise TypeError("step must be an int")
        if step <= 0:
            raise ValueError("step must be > 0")
        return max(self.numbers) + step

    def __get_slice(self, i: slice):
        """Get a new NumberedObjectCollection over a slice of numbers

        This method implements usage like:
            >>> NumberedObjectCollection[1:49:12]
            [1, 13, 25, 37, 49]

        Any ``slice`` object may be passed.
        The indices are the object numbers.
        Because MCNP numbered objects start at 1, so do the indices.
        They are effectively 1-based and endpoint-inclusive.

        :rtype: NumberedObjectCollection
        """
        rstep = i.step if i.step is not None else 1
        rstart = i.start
        rstop = i.stop
        if rstep < 0:  # Backwards
            if rstart is None:
                rstart = max(self.numbers)
            if rstop is None:
                rstop = min(self.numbers)
            rstop -= 1
        else:  # Forwards
            if rstart is None:
                rstart = 0
            if rstop is None:
                rstop = max(self.numbers)
            rstop += 1
        numbered_objects = []
        for num in range(rstart, rstop, rstep):
            obj = self.get(num)
            if obj is not None:
                numbered_objects.append(obj)
        # obj_class is always implemented in child classes.
        return type(self)(numbered_objects)

    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.__get_slice(i)
        elif not isinstance(i, int):
            raise TypeError("index must be an int or slice")
        ret = self.get(i)
        if ret is None:
            raise KeyError(f"Object with number {i} not found in {type(self)}")
        return ret

    def __delitem__(self, idx):
        if not isinstance(idx, int):
            raise TypeError("index must be an int")
        obj = self[idx]
        self.__num_cache.pop(obj.number, None)
        self._objects.remove(obj)

    def __setitem__(self, key, newvalue):
        if not isinstance(key, int):
            raise TypeError("index must be an int")
        self.append(newvalue)

    def __len__(self):
        return len(self._objects)

    def __iadd__(self, other):
        self.extend(other)
        return self

    def __contains__(self, other):
        return other in self._objects

    def get(self, i: int, default=None) -> (Numbered_MCNP_Object, None):
        """
        Get ``i`` if possible, or else return ``default``.

        :param i: number of the object to get, not it's location in the internal list
        :type i: int
        :param default: value to return if not found
        :type default: object

        :rtype: Numbered_MCNP_Object
        """
        try:
            ret = self.__num_cache[i]
            if ret.number == i:
                return ret
        except KeyError:
            pass
        for obj in self._objects:
            if obj.number == i:
                self.__num_cache[i] = obj
                return obj
        return default

    def keys(self) -> typing.Generator[int, None, None]:
        """
        Get iterator of the collection's numbers.

        :rtype: int
        """
        for o in self._objects:
            yield o.number

    def values(self) -> typing.Generator[Numbered_MCNP_Object, None, None]:
        """
        Get iterator of the collection's objects.

        :rtype: Numbered_MCNP_Object
        """
        for o in self._objects:
            yield o

    def items(
        self,
    ) -> typing.Generator[typing.Tuple[int, Numbered_MCNP_Object], None, None]:
        """
        Get iterator of the collections (number, object) pairs.

        :rtype: tuple(int, MCNP_Object)
        """
        for o in self._objects:
            yield o.number, o


class NumberedDataObjectCollection(NumberedObjectCollection):
    def __init__(self, obj_class, objects=None, problem=None):
        self._last_index = None
        if problem and objects:
            try:
                self._last_index = problem.data_inputs.index(objects[-1])
            except ValueError:
                pass
        super().__init__(obj_class, objects, problem)

    def append(self, obj, insert_in_data=True):
        """Appends the given object to the end of this collection.

        :param obj: the object to add.
        :type obj: Numbered_MCNP_Object
        :param insert_in_data: Whether to add the object to the linked problem's data_inputs.
        :type insert_in_data: bool
        :raises NumberConflictError: if this object has a number that is already in use.
        """
        super().append(obj)
        if self._problem:
            if self._last_index:
                index = self._last_index
            elif len(self) > 0:
                try:
                    index = self._problem.data_inputs.index(self._objects[-1])
                except ValueError:
                    index = len(self._problem.data_inputs)
            else:
                index = len(self._problem.data_inputs)
            if insert_in_data:
                self._problem.data_inputs.insert(index + 1, obj)
            self._last_index = index + 1

    def __delitem__(self, idx):
        if not isinstance(idx, int):
            raise TypeError("index must be an int")
        obj = self[idx]
        super().__delitem__(idx)
        if self._problem:
            self._problem.data_inputs.remove(obj)

    def remove(self, delete):
        """
        Removes the given object from the collection.

        :param delete: the object to delete
        :type delete: Numbered_MCNP_Object
        """
        super().remove(delete)
        if self._problem:
            self._problem.data_inputs.remove(delete)

    def pop(self, pos=-1):
        """
        Pop the final items off of the collection

        :param pos: The index of the element to pop from the internal list.
        :type pos: int
        :return: the final elements
        :rtype: Numbered_MCNP_Object
        """
        if not isinstance(pos, int):
            raise TypeError("The index for popping must be an int")
        obj = self._objects.pop(pos)
        super().pop(pos)
        if self._problem:
            self._problem.data_inputs.remove(obj)
        return obj

    def clear(self):
        """
        Removes all objects from this collection.
        """
        if self._problem:
            for obj in self._objects:
                self._problem.data_inputs.remove(obj)
        self._last_index = None
        super().clear()
